@@ -1,11 +1,11 @@
 /*======================================================================
 
    xmix: X interface to the Sound Blaster mixer.
-   [ This file is a part of SBlast-BSD-1.4 ]
+   [ This file is a part of SBlast-BSD-1.5 ]
 
    Steve Haehnichen <shaehnic@ucsd.edu>
  
-   $Id: xmix.c,v 1.4 1992/06/12 22:33:39 steve Exp steve $
+   xmix.c,v 1.5 1992/09/14 03:17:21 steve Exp
 
    Copyright (C) 1992 Steve Haehnichen.
 
@@ -23,7 +23,11 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
- * $Log: xmix.c,v $
+ * xmix.c,v
+ * Revision 1.5  1992/09/14  03:17:21  steve
+ * Released with driver v1.5.
+ * Converted over to stereo CD levels.
+ *
  * Revision 1.4  1992/06/12  22:33:39  steve
  * Much better.
  * Moved many resources into source code.
@@ -113,7 +117,8 @@
 			 RootWindowOfScreen (XtScreen (topLevel)),\
 			 bits, width, height)
 #define SCROLLBAR_RES	 XtNwidth, 23, XtNheight, 100
-
+#define CENTER(widg)	 XtVaSetValues (widg, XtNwidth,\
+					centering_width, NULL)
 
 struct slider_info		/* Client data structure for faders */
 {
@@ -136,23 +141,25 @@ Pixmap circle_on_pixmap, circle_off_pixmap;
 Pixmap square_empty_pixmap, square_with_x_pixmap;
 Widget topLevel, quit, whole, buttons, sliders;
 Widget master_l, master_r, line_l, line_r, dsp_l, dsp_r;
-Widget fm_l, fm_r, cd_level, mic_level;
+Widget fm_l, fm_r, cd_l, cd_r, mic_level;
 Widget master_form, line_form, dsp_form, fm_form, cd_form, mic_form;
 Widget master_label, line_label, dsp_label, fm_label, cd_label, mic_label;
-Widget master_lock, line_lock, dsp_lock, fm_lock;
-struct lock_info master_lock_info, line_lock_info, dsp_lock_info, fm_lock_info;
+Widget master_lock, line_lock, dsp_lock, fm_lock, cd_lock;
+struct lock_info master_lock_info, line_lock_info;
+struct lock_info dsp_lock_info, fm_lock_info, cd_lock_info;
 Widget channels, channels_form;
 Widget filters, out_filt, in_filt, filt_freq, filters_label;
 Widget out_filt_label, in_filt_label;
 Widget sources, line_src, mic_src, cd_src, source_label;
 Widget line_src_label, mic_src_label, cd_src_label;
+int centering_width;
 int mixer_fd;
 struct sb_mixer_params params;
 struct sb_mixer_levels levels;
 struct slider_info master_info_l, line_info_l, dsp_info_l, fm_info_l;
 struct slider_info master_info_r, line_info_r, dsp_info_r, fm_info_r;
-struct slider_info cd_info, mic_info;
-FLAG master_linked, line_linked, dsp_linked, fm_linked;
+struct slider_info cd_info_l, cd_info_r, mic_info;
+FLAG master_linked, line_linked, dsp_linked, fm_linked, cd_linked;
 
 void install_pixmaps (void);
 void sync_display (void);
@@ -174,6 +181,7 @@ void
 main (int argc, char **argv)
 {
   XtAppContext    app_context;
+  int scroll_sep;
   
   topLevel = XtVaAppInitialize (&app_context,
      "XMix",			/* Application class */
@@ -200,7 +208,8 @@ main (int argc, char **argv)
   master_info_r.volume = &levels.master.r;
   master_info_l.linked = master_info_r.linked = &master_linked;
   master_lock = MW ("master_lock", commandWidgetClass, master_form,
-		    XtNfromVert, master_l, NULL);
+		    XtNfromVert, master_l,
+		    NULL);
   master_lock_info.linked = &master_linked;
   master_lock_info.pair = &levels.master;
     
@@ -265,10 +274,20 @@ main (int argc, char **argv)
 		XtNfromHoriz, fm_form, NULL);
   cd_label =  MW ("cd_label", labelWidgetClass, cd_form,
 		  XtNlabel, "CD", NULL);
-  cd_level = MW ("cd_level", scrollbarWidgetClass, cd_form, SCROLLBAR_RES,
-		 XtNfromVert, cd_label, NULL);
-  cd_info.linked = FALSE;
-  cd_info.volume = &levels.cd;
+  cd_l = MW ("cd_l", scrollbarWidgetClass, cd_form, SCROLLBAR_RES,
+	     XtNfromVert, cd_label, NULL);
+  cd_r = MW ("cd_r", scrollbarWidgetClass, cd_form, SCROLLBAR_RES,
+	     XtNfromVert, cd_label,
+	     XtNfromHoriz, cd_l, NULL);
+  cd_lock = MW ("cd_lock", commandWidgetClass, cd_form,
+		XtNfromVert, cd_l, NULL);
+  cd_info_l.other_side = cd_r;
+  cd_info_r.other_side = cd_l;
+  cd_info_l.volume = &levels.cd.l;
+  cd_info_r.volume = &levels.cd.r;
+  cd_info_l.linked = cd_info_r.linked = &cd_linked;
+  cd_lock_info.linked = &cd_linked;
+  cd_lock_info.pair = &levels.cd;
 
   mic_form = MW ("mic_form", formWidgetClass, sliders,
 		 XtNfromHoriz, cd_form, NULL);
@@ -334,6 +353,17 @@ main (int argc, char **argv)
 		     XtNfromVert, mic_src,
 		     XtNfromHoriz, cd_src, NULL);
 
+  XtVaGetValues (mic_level, XtNwidth, &centering_width, NULL);
+  CENTER (mic_label);
+  XtVaGetValues (master_form, XtNhorizDistance, &scroll_sep, NULL);
+  XtVaGetValues (master_l, XtNwidth, &centering_width, NULL);
+  centering_width = centering_width * 2 + scroll_sep;
+  CENTER (master_label);
+  CENTER (line_label);
+  CENTER (dsp_label);
+  CENTER (fm_label);
+  CENTER (cd_label);
+
   XtAddEventHandler (topLevel, EnterWindowMask, FALSE,
 		     (XtEventHandler) sync_display, NULL);
   XtAddCallback (quit, XtNcallback, Quit, 0 /* client_data */ );
@@ -353,8 +383,10 @@ main (int argc, char **argv)
   XtAddCallback (fm_l, XtNscrollProc, Handle_slam_slider, 0);
   XtAddCallback (fm_r, XtNjumpProc, Handle_slider, &fm_info_r);
   XtAddCallback (fm_r, XtNscrollProc, Handle_slam_slider, 0);
-  XtAddCallback (cd_level, XtNjumpProc, Handle_slider, &cd_info);
-  XtAddCallback (cd_level, XtNscrollProc, Handle_slam_slider, 0);
+  XtAddCallback (cd_l, XtNjumpProc, Handle_slider, &cd_info_l);
+  XtAddCallback (cd_l, XtNscrollProc, Handle_slam_slider, 0);
+  XtAddCallback (cd_r, XtNjumpProc, Handle_slider, &cd_info_r);
+  XtAddCallback (cd_r, XtNscrollProc, Handle_slam_slider, 0);
   XtAddCallback (mic_level, XtNjumpProc, Handle_slider, &mic_info);
   XtAddCallback (mic_level, XtNscrollProc, Handle_slam_slider, 0);
   XtAddCallback (channels, XtNcallback, Handle_toggle, &params.dsp_stereo);
@@ -367,6 +399,7 @@ main (int argc, char **argv)
   XtAddCallback (master_lock, XtNcallback, Handle_lock, &master_lock_info);
   XtAddCallback (line_lock, XtNcallback, Handle_lock, &line_lock_info);
   XtAddCallback (fm_lock, XtNcallback, Handle_lock, &fm_lock_info);
+  XtAddCallback (cd_lock, XtNcallback, Handle_lock, &cd_lock_info);
   XtAddCallback (dsp_lock, XtNcallback, Handle_lock, &dsp_lock_info);
 
   install_pixmaps ();
@@ -389,6 +422,7 @@ main (int argc, char **argv)
   line_linked = (levels.line.l == levels.line.r);
   dsp_linked = (levels.voc.l == levels.voc.r);
   fm_linked = (levels.fm.l == levels.fm.r);
+  cd_linked = (levels.cd.l == levels.cd.r);
 
   /*
    * Update the lock bitmaps to reflect linking
@@ -397,6 +431,7 @@ main (int argc, char **argv)
   sync_lock (line_linked, line_lock);
   sync_lock (dsp_linked, dsp_lock);
   sync_lock (fm_linked, fm_lock);
+  sync_lock (cd_linked, cd_lock);
 
   XtRealizeWidget (topLevel);	/* Action! */
   XtAppMainLoop (app_context);	/* Loop for events */
@@ -453,7 +488,8 @@ sync_sliders (void)
   XawScrollbarSetThumb (fm_r, PORTION(levels.fm.r), -1.0);
   XawScrollbarSetThumb (line_l, PORTION(levels.line.l), -1.0);
   XawScrollbarSetThumb (line_r, PORTION(levels.line.r), -1.0);
-  XawScrollbarSetThumb (cd_level, PORTION(levels.cd), -1.0);
+  XawScrollbarSetThumb (cd_l, PORTION(levels.cd.l), -1.0);
+  XawScrollbarSetThumb (cd_r, PORTION(levels.cd.r), -1.0);
   XawScrollbarSetThumb (mic_level, (1.0 - (float)levels.mic / 7.0), -1.0);
 }
 
@@ -464,6 +500,7 @@ sync_lock (int flag, Widget w)
       XtVaSetValues (w, XtNbitmap, locked_pixmap, NULL);
   else
       XtVaSetValues (w, XtNbitmap, unlocked_pixmap, NULL);
+  CENTER (w);
 }
 
 void
