@@ -1,11 +1,11 @@
 /*======================================================================
 
    Device driver for the Creative Labs Sound Blaster card.
-   [ This file is a part of SBlast-BSD-1.4 ]
+   [ This file is a part of SBlast-BSD-1.5 ]
 
    Steve Haehnichen <shaehnic@ucsd.edu>
  
-   $Id: sb_driver.c,v 1.29 1992/06/13 01:46:43 steve Exp steve $
+   sb_driver.c,v 2.2 1992/09/14 03:09:22 steve Exp
 
    Copyright (C) 1992 Steve Haehnichen.
 
@@ -23,9 +23,29 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
- * $Log: sb_driver.c,v $
+ * sb_driver.c,v
+ * Revision 2.2  1992/09/14  03:09:22  steve
+ * Released as distribution v1.5.
+ * Mostly stable, still lacking stereo recording.
+ *
+ * Revision 2.1  1992/08/02  21:54:59  steve
+ * Added select handler and streamlined a few things.
+ * Now allows read/write on a single open.
+ *
+ * Revision 2.0  1992/07/13  01:25:14  steve
+ * Promoting to v2.0 with start of CVS management.
+ *
+ * Revision 1.6  1992/07/13  01:14:21  steve
+ * Fixed version printing
+ *
+ * Revision 1.5.1.1  1992/07/13  00:39:30  steve
+ * Stable double-buffering Sound Blaster driver.
+ *
+ * Revision 1.30  1992/06/20  04:41:29  steve
+ * About to re-write buffering.
+ *
  * Revision 1.29  1992/06/13  01:46:43  steve
- * Released in SBlast-BSD-1.4
+ * Released in SBlast-BSD-1.5
 
 ======================================================================*/
 
@@ -54,29 +74,12 @@
 #  define DEBUG
 #endif
 
-/*
- * Here's the debugging macro I use here and there, so I can turn
- * them all on or off in one place.
- */
-#ifdef DEBUG
-#  define DPRINTF(x)	printf x
-#else
-#  define DPRINTF(x)
-#endif
-
-/*
- * If this is defined, DSP ioctls will be allowed on any device.
- * This is a debugging thing, and should be disabled before final
- * installation.  (It allows me to send a DSP_RESET to the SB
- * while the DSP device is "busy" making horrible noises.)
- */
-/* #define 	ALWAYS_ALLOW_DSP_IOCTLS */
-
 /*
  * The Mixer is an odd deal.  I think Mixer ioctls should be
  * allowed on any device.  It's very safe.
  * If all your users complain that other users are messing up the
- * mixer while they are playing, well, undef it then. :-)
+ * mixer while they are playing, well, undef this and make
+ * mixer open's exclusive.
  */
 #define 	ALWAYS_ALLOW_MIXER_IOCTLS
 
@@ -91,41 +94,43 @@
  * If I can read the speed registers, then maybe I should just use
  * the current card setting.  Does it matter?
  * (43478 is the closest we can get to 44 KHz)
+ * Make sure this is a valid speed for your card!
  */
 #define		INITIAL_DSP_SPEED	43478
 
-/*
- * Define this if you consider ADC overruns to be an error, and
- * want to abort the read() call.  If you are making real recordings,
- * this is usually the case, or you will have gaps in the sound.
- * (This shouldn't happen.  Most apps can keep up with 45K/sec.)
- * For real-time use, overruns should be ignored.
- * This looks like an ioctl-toggle candidate to me. 
- */
-#define 	ERROR_ON_OVERRUN
-
 /*
  * SB Device minor numbers
  * I have copied the numbers used in Brian Smith's ISC Unix driver,
  * with the faint hope of future compatibility.
+ * Oh, make sure this is a valid speed for your card!
  * The listed device names are simply the ones that I am using.
  * Note that there is absolutely no CMS support in this driver.
  */
-#define SB_CMS_NUM  	0	/* /dev/sb_cms   CMS square wave chips */
-#define SB_FM_NUM    	1	/* /dev/sb_fm    FM synthesis chips */
-#define SB_DSP_NUM    	2	/* /dev/sb_dsp   DSP DAC/ADC */
-#define SB_MIDI_NUM   	3	/* /dev/sb_midi  MIDI port */
+#define SB_CMS_NUM  	0	/* /dev/sb_cms   CMS square wave chips   */
+#define SB_FM_NUM    	1	/* /dev/sb_fm    FM synthesis chips      */
+#define SB_DSP_NUM    	2	/* /dev/sb_dsp   DSP DAC/ADC             */
+#define SB_MIDI_NUM   	3	/* /dev/sb_midi  MIDI port               */
 #define SB_MIXER_NUM  	4	/* /dev/sb_mixer Non-exclusive-use Mixer */
-#define SB_BIGGEST_MINOR 4	/* Highest minor number used. */
+#define SB_BIGGEST_MINOR 4	/* Highest minor number used.            */
 
 /*
  * This is the "channel" and IRQ used for the FM timer functions.
  * Since I can't seem to get one to trigger yet,
  * I don't know what they should be.  (see fm_delay() comments)
  */
-#define TIMER_CHANNEL		BOTH
+#define TIMER_CHANNEL		FM_BOTH
 #define FM_TIMER_IRQ		8
 
+/*
+ * Here's the debugging macro I use here and there, so I can turn
+ * them all on or off in one place.
+ */
+#ifdef DEBUG
+#  define DPRINTF(x)	printf x
+#else
+#  define DPRINTF(x)
+#endif
+
 /* Unix interrupt priority.
    I set it to just above Hard Disk priority to keep things smooth. */
 #define INTR_PRIORITY	(PRIBIO - 5)
@@ -135,11 +140,26 @@
 #define ON   1
 #define OFF  0
 
+/* These are used to indicate what the DSP last did,
+   and can safely do again without preparation */
+#define NONE_READY	0
+#define READ_READY	1
+#define WRITE_READY	2
+
+/* These are the possible values for argument 'rw' in the select handlers.
+   Shouldn't they be in an include file somewhere? */
+#define SELECT_READ	1
+#define SELECT_WRITE	2
+
 /* Number of "cycles" in one second.  Shouldn't this be somewhere else? */
 #define HZ 		100
 
 #define TIMEOUT		(10 * HZ) /* FM interrupt patience in clock ticks */
 
+/* This sets dsp.timeout to an appropriate number of ticks */
+#define COMPUTE_DSP_TIMEOUT \
+  { dsp.timeout = HZ + (HZ * dsp.bufsize / dsp.speed); }
+
 /* Semaphore return codes.  This tells the reason for the wakeup(). */
 #define WOKEN_BY_INTERRUPT	1
 #define WOKEN_BY_TIMEOUT	2
@@ -154,6 +174,9 @@
 
 enum { PLAY, RECORD };		/* DSP sampling directions */
 
+#define LOWBYTE(x)	((x) & 0x00ff)
+#define HIBYTE(x)	(((x) & 0xff00) >> 8)
+
 /* Handy macro from Brian Smith's/Lance Norskog's driver */
 #define crosses_64k_boundary(a,b) ((((int)(a) & 0xffff) + (b)) > 0x10000)
 
@@ -172,8 +195,9 @@ enum { PLAY, RECORD };		/* DSP sampling directions */
  * 64K is the biggest DSP_BUF_SIZE you can have.
  * Smaller buffer sizes make aborts and small reads more responsive.
  * For real-time stuff, perhaps ~10K would be better.
+ * Note that you can now change this to any smaller value with the ioctl
+ * DSP_IOCTL_BUFSIZE, so you are probably best off with a 64K limit.
  */
-#define DSP_BUF_SIZE		(64 * 1024)
 static char memory_chunk[3 * 64 * 1024]; /* Be careful!  See above */
 
 /* This structure will be instantiated as a single global holder
@@ -195,8 +219,11 @@ struct sb_dsp_type
   int	used[2];		/* Buffer bytes used by read/write */
   BYTE  active;			/* The buffer currently engaged in DMA */
   BYTE  hi;			/* The buffer being filled/emptied by user */
-  FLAG	first_block;		/* True for the first DAC block */
-  FLAG  dont_block;		/* FNDELAY flag sets this, clear otherwise */
+  int	bufsize;		/* DMA transfer buffer size */
+  int	last_xfer_size;		/* How many bytes we last transfered */
+  int   overrun_errno;		/* errno code for a DSP overrun (0 to ignore)*/
+  int	ver_major, ver_minor;	/* Major and Minor version number */
+  BYTE	state;			/* Either READ_READY, WRITE_READY, or 0 */
 };
 static struct sb_dsp_type dsp;	/* 'dsp' structure used everywhere */
 
@@ -211,6 +238,12 @@ struct sb_status_type
   FLAG  alive;                  /* Card present? */
   unsigned int addr;		/* Sound Blaster card address */
   int  *wake[NUM_UNITS];	/* What to wakeup on interrupt */
+
+  /* Select handler stuff, one entry for each "device" */
+  struct proc *selects[SB_BIGGEST_MINOR];
+  FLAG sel_coll[SB_BIGGEST_MINOR];
+  struct proc *sig_proc[SB_BIGGEST_MINOR];
+  int flags[SB_BIGGEST_MINOR];	/* fcntl flags */
 };
 static struct sb_status_type status; /* Global current status */
 
@@ -233,25 +266,29 @@ int sb_close (dev_t dev, int flags);
 int sb_ioctl (dev_t dev, int cmd, caddr_t arg, int mode);
 int sb_read (int dev, struct uio *uio);
 int sb_write (int dev, struct uio *uio);
+int sb_select (int dev, int rw);
 static void sb_sendb (unsigned select_addr, BYTE reg,
 		      unsigned data_addr, BYTE value);
 static int dsp_reset (void);
-static int dsp_open (int flags);
-static int dsp_close (int flags);
+static int dsp_get_version (void);
+static int dsp_open (void);
+static int dsp_close (void);
 static int dsp_set_speed (int *speed);
 static int dsp_command (BYTE val);
 static int dsp_ioctl (int cmd, caddr_t arg);
 static int dsp_set_voice (int on);
+static int dsp_set_bufsize (int *newsize);
 static int dsp_flush_dac (void);
 static int dsp_set_compression (int mode);
 static int dsp_set_stereo (FLAG on);
 static int dsp_write (struct uio *uio);
 static int dsp_read (struct uio *uio);
+static int dsp_select (int rw);
 static void dsp_find_buffers (void);
 inline static void dsp_dma_start (int dir);
 inline static void dsp_next_write (void);
 inline static void dsp_next_read (void);
-static int fm_open (int flags);
+static int fm_open (void);
 static int fm_close (void);
 static int fm_reset (void);
 static int fm_send (int channel, unsigned char reg, unsigned char val);
@@ -270,11 +307,12 @@ static int mixer_set_levels (struct sb_mixer_levels *l);
 static int mixer_set_params (struct sb_mixer_params *p);
 static int mixer_get_levels (struct sb_mixer_levels *l);
 static int mixer_get_params (struct sb_mixer_params *params);
-static int midi_open (int flags);
-static int midi_close ();
+static int midi_open (void);
+static int midi_close (void);
 static int midi_read (struct uio *uio);
 static int midi_write (struct uio *uio);
 static int midi_ioctl (int cmd, caddr_t arg);
+static int midi_select (int rw);
 
 
 /*
@@ -331,14 +369,32 @@ sb_attach (struct isa_dev *dev)
 
   take_dev_irq (dev);		/* Set up DSP interrupt */
 
-  printf ("sb: Sound Blaster installed. (Port 0x%X, IRQ %d, DMA %d)\n",
+  dsp_get_version ();
+  printf ("sb: Sound Blaster v%d.%02d installed. ", /* %02d is hosed? */
+	  dsp.ver_major, dsp.ver_minor);
+  printf ("(Port 0x%x, IRQ %d, DMA %d)\n",
 	  status.addr, dev->dev_pic, SB_DMA_CHAN);
+  
+  if (dsp.ver_major != EXPECTED_DSP_VER)
+    {
+      printf ("sb: SB Driver is MISCONFIGURED.\n"
+	      "sb: You do not have a DSP version %d.x.\n", EXPECTED_DSP_VER);
+      status.alive = FALSE;
+    }
+
   status.fm_in_use = 0;
   status.dsp_in_use = 0;
   dsp_find_buffers();
   for (i = 0; i < NUM_UNITS; i++)
     status.wake[i] = NULL;
 
+  for (i = 0; i <= SB_BIGGEST_MINOR; i++)
+    {
+      status.sig_proc[i] = NULL;
+      status.selects[i] = NULL;
+      status.sel_coll[i] = 0;
+    }
+
   /*
    * These are default startup settings.
    * I'm wondering if I should just leave the card in the same
@@ -348,7 +404,8 @@ sb_attach (struct isa_dev *dev)
   dsp.compression = PCM_8;
   dsp.speed = INITIAL_DSP_SPEED;
   dsp_set_stereo (FALSE);
-
+  dsp.bufsize = DSP_MAX_BUFSIZE;
+  /* mixer_reset(); */
   return (ESUCCESS);
 }
 
@@ -361,6 +418,7 @@ sb_attach (struct isa_dev *dev)
 int
 sb_intr (int unit)
 {
+  int i;
   DPRINTF (("sb_intr(): Got interrupt on unit %d\n", unit));
 
   /*
@@ -371,12 +429,39 @@ sb_intr (int unit)
   if (dsp.cont_xfer && unit == DSP_UNIT)
     (*dsp.cont_xfer)();
 
+  /*
+   * If any selects are pending on any devices, wake them up
+   * and have them check for data ready.  Also send out ASYNC
+   * signals to anyone who asked for them.
+   * Of course, this makes the blatant assumption that only one
+   * interrupt-generating device will be open at any time, since
+   * we can't figure out what caused the SB to interrupt.
+   * (Ahhh, I love that ISA bus..)
+   */
+  for (i = 0; i <= SB_BIGGEST_MINOR; i++)
+    {
+      if (status.selects[i])
+	{
+	  DPRINTF (("Waking up selecters..\n"));
+	  selwakeup (status.selects[i], status.sel_coll[i]);
+	  status.selects[i] = 0;
+	  status.sel_coll[i] = FALSE;
+	}
+      /*
+       * I'm not sure if this is the right time to send out signals.
+       * Perhaps more checking should be done to make sure the process
+       * is really interested in this interrupt.  I don't use ASYNC much.
+       */
+      if (status.sig_proc[i])
+	psignal (status.sig_proc[i], SIGIO);
+    }
+  
   /*
    * If this Unit is expecting an interrupt, then set the semaphore
    * and wake up the sleeper.  Otherwise, we probably weren't expecting
    * this interrupt, so just ignore it.  This also happens when two
    * interrupts are acked in a row, without sleeping in between.
-   * Not really a problem, but there should be a clean way to fix it.
+   * Not really a problem, but there should be a cleaner way to fix it.
    */
   if (status.wake[unit] != NULL)
     {
@@ -386,7 +471,7 @@ sb_intr (int unit)
     }
   else
     {
-      DPRINTF (("sb_intr(): An ignored interrupt!\n"));
+      DPRINTF (("sb_intr(): An ignored interrupt.\n"));
     }
   return (ESUCCESS);
 }
@@ -400,14 +485,20 @@ sb_open (dev_t dev, int flags)
 
   DPRINTF (("open(): dev = 0x%X, flags = 0x%X\n", dev, flags));
 
+  /* Remember the flags for future reference */
+  status.flags[minor(dev)] = flags;
+  status.sig_proc[minor(dev)] = NULL;
+  status.selects[minor(dev)] = NULL;
+  status.sel_coll[minor(dev)] = 0;
+
   switch (minor (dev))
     {
     case SB_DSP_NUM:
-      return (dsp_open (flags));
+      return (dsp_open ());
     case SB_FM_NUM:
-      return (fm_open (flags));
+      return (fm_open ());
     case SB_MIDI_NUM:
-      return (midi_open (flags));
+      return (midi_open ());
     case SB_MIXER_NUM:
       return (ESUCCESS);
       
@@ -427,7 +518,7 @@ sb_close (dev_t dev, int flags)
   switch (minor (dev))
     {
     case SB_DSP_NUM:
-      return (dsp_close (flags));
+      return (dsp_close ());
     case SB_FM_NUM:
       return (fm_close ());
     case SB_MIDI_NUM:
@@ -463,15 +554,18 @@ sb_ioctl (dev_t dev, int cmd, caddr_t arg, int mode)
     case MIXER_IOCTL_READ_PARAMS:
       return (mixer_ioctl (cmd, arg));
 #endif
-
-#ifdef ALWAYS_ALLOW_DSP_IOCTLS	/* Mostly for easy emergency aborts */
-    case DSP_IOCTL_FLUSH:
-    case DSP_IOCTL_RESET:
-    case DSP_IOCTL_SPEED:
-    case DSP_IOCTL_STEREO:
-      return (dsp_ioctl (cmd, arg));
-#endif
-
+      /*
+       * No support for these yet.  Soon..
+       * For now, you just have to open the device with the right flags.
+       * As soon as I figure out fcntl() calls, this will work.
+       */
+    case F_SETFL:
+      DPRINTF (("Got F_SETFL ioctl.\n"));
+      return (EINVAL);
+    case F_GETFL:
+      DPRINTF (("Got F_GETFL ioctl.\n"));
+      return (EINVAL);
+      
     default:
       switch (minor (dev))
 	{
@@ -489,6 +583,7 @@ sb_ioctl (dev_t dev, int cmd, caddr_t arg, int mode)
 	  return (EINVAL);
 	}
     }      
+  return (ESUCCESS);
 }
 
 /* Multiplexes a read to the different devices */
@@ -541,6 +636,24 @@ sb_write (int dev, struct uio *uio)
     }
 }
 
+int
+sb_select (int dev, int rw)
+{
+  if (!status.alive)
+    return (ENODEV);
+  DPRINTF (("sb: sb_select: dev = 0x%X\n", dev));
+  switch (minor (dev))
+    {
+    case SB_DSP_NUM:
+      return (dsp_select (rw));
+    case SB_MIDI_NUM:
+      return (midi_select (rw));
+    default:
+      printf ("sb: sb_select() on bad dev.\n");
+      return (1);
+    }
+}
+
 /*
  * Send one byte to the named Sound Blaster register.
  * The SBDK recommends 3.3 microsecs after an address write,
@@ -615,6 +728,27 @@ wait_for_intr (int unit, int patience, int priority)
 /*
  * DSP functions.
  */
+static int
+dsp_select (int rw)
+{
+  int s = splhigh();
+  DPRINTF (("sb: dsp_select() called with rw = %d.\n", rw));
+  if ((dsp.cont_xfer == NULL)
+      || ((rw == SELECT_WRITE) && (dsp.active == dsp.hi))
+      || ((rw == SELECT_READ) && (dsp.full[dsp.hi] == TRUE)))
+    {
+      splx (s);
+      return (1);
+    }
+  if (status.selects[SB_DSP_NUM])
+    status.sel_coll[SB_DSP_NUM] = TRUE;
+  else
+    status.selects[SB_DSP_NUM] = (struct proc *) current_thread(); /* eeyuck */
+  
+  splx (s);
+  return 0;
+}
+
 static int
 dsp_ioctl (int cmd, caddr_t arg)
 {
@@ -625,16 +759,21 @@ dsp_ioctl (int cmd, caddr_t arg)
 	return (ESUCCESS);
       else
 	return (EIO);
+    case DSP_IOCTL_OVERRUN_ERRNO:
+      dsp.overrun_errno = (*(int *)arg);
+      return (ESUCCESS);
+    case DSP_IOCTL_BUFSIZE:
+      return (dsp_set_bufsize ((int *)arg));
     case DSP_IOCTL_SPEED:
       return (dsp_set_speed ((int *)arg));
     case DSP_IOCTL_VOICE:
-      return (dsp_set_voice (*(FLAG*)arg));
+      return (dsp_set_voice (*(FLAG *)arg));
     case DSP_IOCTL_FLUSH:
       return (dsp_flush_dac ());
     case DSP_IOCTL_COMPRESS:
-      return (dsp_set_compression (*(BYTE *) arg));
+      return (dsp_set_compression (*(BYTE *)arg));
     case DSP_IOCTL_STEREO:
-      return (dsp_set_stereo (*(FLAG *) arg));
+      return (dsp_set_stereo (*(FLAG *)arg));
     default:
       return (EINVAL);
     }
@@ -659,9 +798,10 @@ dsp_reset (void)
   dsp.used[0] = dsp.used[1] = 0; /* This is only for write; see dsp_read() */
   dsp.full[0] = dsp.full[1] = FALSE;
   dsp.hi = dsp.active = 0;
-  dsp.first_block = 1;
+  dsp.state = NONE_READY;
   dsp.error = ESUCCESS;
   dsp.cont_xfer = NULL;
+  dsp.last_xfer_size = -1;
   status.wake[DSP_UNIT] = NULL;
 
   /*
@@ -696,6 +836,37 @@ dsp_reset (void)
     return (GOOD);
 }
 
+static int
+dsp_get_version (void)
+{
+  int i;
+  
+  dsp.ver_major = dsp.ver_minor = 0;
+  dsp_command (GET_DSP_VERSION); /* Request DSP version number */
+  for (i = 10 * DSP_LOOP_MAX; i; i--)
+    {
+      if (inb (DSP_RDAVAIL) & DSP_DATA_AVAIL) /* wait for Data Ready */
+	{
+	  if (dsp.ver_major == 0)
+	    dsp.ver_major = inb (DSP_RDDATA);
+	  else
+	    {
+	      dsp.ver_minor = inb (DSP_RDDATA);
+	      break;		/* All done */
+	    }
+	}
+      tenmicrosec ();
+    }
+
+  if (i)			/* finished in time */
+    return (GOOD);
+  else
+    {
+      printf ("sb: Error reading DSP version number.\n");
+      return (FAIL);
+    }
+}
+
 /* 
  * This finds page-safe buffers for the DSP DMA to use.  A single DMA
  * transfer can never cross a 64K page boundary, so we have to get
@@ -715,16 +886,17 @@ dsp_find_buffers (void)
 					    & ~0xFFFF));
   dsp.buf[1] = dsp.buf[0] + 0x10000;
   
-  if (crosses_64k_boundary (dsp.buf[0], DSP_BUF_SIZE)
-      || crosses_64k_boundary (dsp.buf[1], DSP_BUF_SIZE))
+  if (crosses_64k_boundary (dsp.buf[0], DSP_MAX_BUFSIZE)
+      || crosses_64k_boundary (dsp.buf[1], DSP_MAX_BUFSIZE))
     panic ("sb: Couldn't get good DSP buffers!\n");
 
   dsp.phys_buf[0] = kvtophys(dsp.buf[0]);
   dsp.phys_buf[1] = kvtophys(dsp.buf[1]);
 }
 
+
 /*
- * Start a DMA transfer to/from the DSP.
+ * Start a single-buffer DMA transfer to/from the DSP.
  * This one needs more explaining than I would like to put in comments,
  * so look at the accompanying documentation, which, of course, hasn't
  * been written yet. :-)
@@ -745,10 +917,10 @@ dsp_dma_start (int dir)
   outb (DMA_CLEAR, 0);
   outb (DMA_MODE, (dir == RECORD) ? DMA_MODE_READ : DMA_MODE_WRITE);
   outb (DMA_PAGE, (dsp.phys_buf[dsp.active] & 0xff0000) >> 16);	/* Page */
-  outb (DMA_ADDRESS, dsp.phys_buf[dsp.active] & 0x00ff); /* LSB of address */
-  outb (DMA_ADDRESS, (dsp.phys_buf[dsp.active] & 0xff00) >> 8);
-  outb (DMA_COUNT, count & 0x00ff);
-  outb (DMA_COUNT, (count & 0xff00) >> 8);
+  outb (DMA_ADDRESS, LOWBYTE (dsp.phys_buf[dsp.active]));
+  outb (DMA_ADDRESS, HIBYTE (dsp.phys_buf[dsp.active]));
+  outb (DMA_COUNT, LOWBYTE(count));
+  outb (DMA_COUNT, HIBYTE(count));
   outb (DMA_MASK_REG, DMA_UNMASK);
 
   /*
@@ -763,11 +935,12 @@ dsp_dma_start (int dir)
       DPRINTF (("Starting High-Speed DMA of %d bytes to/from buffer %d.\n",
 		dsp.used[dsp.active], dsp.active));
 
-      if (count != DSP_BUF_SIZE - 1)
+      if (count != dsp.last_xfer_size)
 	{
-	  dsp_command (HIGH_SPEED_SIZE);
-	  dsp_command (count & 0x00ff);
-	  dsp_command ((count & 0xff00) >> 8);
+	  dsp_command (TRANSFER_SIZE);
+	  dsp_command (LOWBYTE (count));
+	  dsp_command (HIBYTE (count));
+	  dsp.last_xfer_size = count;
 	}
       dsp_command (dsp.start_command); /* GO! */
     }
@@ -776,8 +949,8 @@ dsp_dma_start (int dir)
       DPRINTF (("Starting Low-Speed DMA xfer of %d bytes to/from buffer %d.\n",
 		dsp.used[dsp.active], dsp.active));
       dsp_command (dsp.start_command);
-      dsp_command (count & 0x00ff);
-      dsp_command ((count & 0xff00) >> 8); /* GO! */
+      dsp_command (LOWBYTE (count));
+      dsp_command (HIBYTE (count)); /* GO! */
     }
 
   /* This sets up the function to call at the next interrupt: */
@@ -828,15 +1001,13 @@ dsp_next_read (void)
   dsp.active ^= 1;
   if (dsp.full[dsp.active])
     {
-#ifdef ERROR_ON_OVERRUN
       /*
        * An overrun occurs when we fill up two buffers faster than the
        * user is reading them.  Lossage has to occur.  This may or
        * may not be bad.  For recording, it's bad.  For real-time
        * FFTs and such, it's not a real big deal.
        */
-      dsp.error = ERANGE;
-#endif
+      dsp.error = dsp.overrun_errno;
       dsp.cont_xfer = NULL;
     }
   else
@@ -852,20 +1023,33 @@ dsp_next_read (void)
  * When one buffer is done filling, we fill the other one while copying
  * the fresh data to the user.
  * If the user doesn't read fast enough for that DSP speed, we have an
- * overrun.  See above concerning ERROR_ON_OVERRUN.
+ * overrun.  See above concerning overrun_errno.
  */
 static int
 dsp_read (struct uio *uio)
 {
   int bytecount, hunk_size, s, chunk;
 
-  if (dsp.first_block)
+  if (dsp.state != READ_READY)
     {
-      DPRINTF (("Kicking in first_block DSP read..\n"));
+      if (!(status.flags[SB_DSP_NUM] & FREAD))
+	return (EINVAL);
+      dsp_flush_dac ();		/* In case we were writing stuff */
+
       s = splhigh ();
-      dsp.first_block = FALSE;
+      dsp_set_voice (OFF);
+      dsp_set_speed (&dsp.speed); /* Set SB to the current speed. */
+      dsp.state = READ_READY;
+      DPRINTF (("Kicking in first DSP read..\n"));
       dsp.start_command = dsp.hispeed ? HS_ADC_8 : ADC_8;
-      dsp.used[0] = dsp.used[1] = DSP_BUF_SIZE;	/* Start with both empty */
+
+      /* A DSP state reset, without the hardware reset. */
+      dsp.used[0] = dsp.used[1] = dsp.bufsize; /* Start with both empty */
+      dsp.full[0] = dsp.full[1] = FALSE;
+      dsp.hi = dsp.active = 0;
+      dsp.error = ESUCCESS;
+      dsp.last_xfer_size = -1;
+      status.wake[DSP_UNIT] = NULL;
       dsp_dma_start (RECORD);
       splx (s);
     }
@@ -891,7 +1075,7 @@ dsp_read (struct uio *uio)
 
 	  /* Now we give a piece of the buffer to the user */
 	  hunk_size = MIN (uio->uio_iov[chunk].iov_len - bytecount,
-			   DSP_BUF_SIZE - dsp.used[dsp.hi]);
+			   dsp.bufsize - dsp.used[dsp.hi]);
 
 	  DPRINTF (("Copying %d bytes from buffer %d.\n", hunk_size, dsp.hi));
 	  if (copyout (dsp.buf[dsp.hi] + dsp.used[dsp.hi], 
@@ -902,12 +1086,19 @@ dsp_read (struct uio *uio)
 	    }
 	  dsp.used[dsp.hi] += hunk_size;
 
-	  if (dsp.used[dsp.hi] == DSP_BUF_SIZE)
+	  if (dsp.used[dsp.hi] == dsp.bufsize)
 	    {
 	      DPRINTF (("Drained all of buffer %d.\n", dsp.hi));
 	      dsp.full[dsp.hi] = FALSE;
 	      dsp.hi ^= 1;
 	      dsp.used[dsp.hi] = 0;
+	      if (dsp.cont_xfer == NULL) /* We had an overrun recently */
+		{
+		  DPRINTF (("Caught up after overrun, starting DMA..\n"));
+		  s = splhigh ();
+		  dsp_dma_start (RECORD);
+		  splx (s);
+		}
 	    }
 	  bytecount += hunk_size;
 	}			/* While there are bytes left in chunk */
@@ -927,24 +1118,32 @@ dsp_write (struct uio *uio)
 {
   int bytecount, hunk_size, s, chunk;
 
-  /*
-   * If this is the first write() operation for this open(),
-   * then figure out the DSP command to use.
-   * Have to check if it is High Speed, or one of the compressed modes.
-   * If this is the first block of a Compressed sound file,
-   * then we have to set the "Reference Bit" in the dsp.command for the
-   * first block transfer.
-   */
-  if (dsp.first_block)
+  if (dsp.state != WRITE_READY)
     {
-      dsp.first_block = FALSE;
+      if (!(status.flags[SB_DSP_NUM] & FWRITE))
+	return (EINVAL);
+      dsp_reset ();
+      dsp_set_speed (&dsp.speed); /* Set SB back to the current speed. */
+      dsp_set_voice (ON);
+      dsp.state = WRITE_READY;
+    
+      /*
+       * If this is the first write() operation for this open(),
+       * then figure out the DSP command to use.
+       * Have to check if it is High Speed, or one of the compressed modes.
+       * If this is the first block of a Compressed sound file,
+       * then we have to set the "Reference Bit" in the dsp.command for the
+       * first block transfer.
+       */
       if (dsp.hispeed)
-	dsp.start_command = HS_DAC_8;
+	{
+	  dsp.start_command = HS_DAC_8;
+	}
       else
 	{
 	  dsp.start_command = dsp.compression;
 	  if (dsp.compression == ADPCM_4
-	      || dsp.compression == ADPCM_2_6
+	      || dsp.compression == ADPCM_3
 	      || dsp.compression == ADPCM_2)
 	    dsp.start_command |= 1;
 	}
@@ -962,9 +1161,9 @@ dsp_write (struct uio *uio)
 	    return (dsp.error);
 
 	  hunk_size = MIN (uio->uio_iov[chunk].iov_len - bytecount,
-			   DSP_BUF_SIZE - dsp.used[dsp.hi]);
+			   dsp.bufsize - dsp.used[dsp.hi]);
 	  DPRINTF (("Adding %d bytes (%d) to buffer %d.\n",
-		  hunk_size, dsp.used[dsp.hi], dsp.hi));
+		    hunk_size, dsp.used[dsp.hi], dsp.hi));
 	  if (copyin (uio->uio_iov[chunk].iov_base + bytecount,
 		      dsp.buf[dsp.hi] + dsp.used[dsp.hi], hunk_size)) 
 	    {
@@ -973,7 +1172,7 @@ dsp_write (struct uio *uio)
 	    }
 	  dsp.used[dsp.hi] += hunk_size;
 
-	  if (dsp.used[dsp.hi] == DSP_BUF_SIZE)
+	  if (dsp.used[dsp.hi] == dsp.bufsize)
 	    {
 	      s = splhigh ();
 	      dsp.full[dsp.hi] = TRUE;
@@ -1044,47 +1243,33 @@ dsp_flush_dac (void)
       DPRINTF (("Waiting for last DMA(s) to finish.\n"));
       wait_for_intr (DSP_UNIT, dsp.timeout, INTR_PRIORITY);
     }
+  dsp.used[0] = dsp.used[1] = 0;
   splx (s);
   return (ESUCCESS);
 }
 
 
 static int
-dsp_open (int flags)
+dsp_open (void)
 {
   if (status.dsp_in_use)
     return (EBUSY);
 
-  /* Hardware does not support simultaneous dma input and output */
-  if ((flags & FREAD) && (flags & FWRITE))
-    return (EINVAL);
-
+  status.selects[SB_DSP_NUM] = NULL;
+  status.sel_coll[SB_DSP_NUM] = FALSE;
   status.dsp_in_use = TRUE;
-  dsp.dont_block = flags & O_NDELAY;
-  dsp_reset ();			/* Resets card and inits variables */
-  dsp_set_speed (&dsp.speed);	/* Set SB back to the current speed. */
 
-  /*
-   * This next step is very important!
-   * I learned the hard way that everything will LOOK like it is working,
-   * but only silence on the play or record.
-   */
-  if (flags & FREAD)
-    dsp_set_voice (OFF);
-  else
-    dsp_set_voice (ON);
+  /* Take note of the process if they want asynchronous signals */
+  if (status.flags[SB_DSP_NUM] & FASYNC)
+    status.sig_proc[SB_DSP_NUM] = u.u_procp;
   
-  /*
-   * In case we do any High-Speed transfers, this sets the transfer
-   * size.  We only need to set it when it changes.  See dsp_dma_start()
-   */
-  dsp_command (HIGH_SPEED_SIZE);
-  dsp_command ((DSP_BUF_SIZE - 1) & 0x00ff);
-  dsp_command (((DSP_BUF_SIZE - 1) & 0xff00) >> 8);
+  dsp_reset ();			/* Resets card and inits variables */
+  dsp.overrun_errno = EIO;	/* Until we are told otherwise */
+  dsp.state = NONE_READY;	/* Not ready for either reading or writing */
 
 #ifdef FULL_DEBUG
   /* Stuff buffers with loud garbage so we can hear/see leaks. */
-  for (i = 0; i < DSP_BUF_SIZE; i++)
+  for (i = 0; i < dsp.bufsize; i++)
     dsp.buf[0][i] = dsp.buf[1][i] = i & 0xff;
 #endif
   return (ESUCCESS);
@@ -1092,12 +1277,12 @@ dsp_open (int flags)
 
 
 static int
-dsp_close (int flags)
+dsp_close (void)
 {
   if (status.dsp_in_use)
     {
       /* Wait for any last write buffers to empty  */
-      if (flags & FWRITE)
+      if (dsp.state == WRITE_READY)
 	dsp_flush_dac ();
       dsp_reset ();
       status.dsp_in_use = FALSE;
@@ -1125,17 +1310,16 @@ dsp_set_speed (int *speed)
 {
   BYTE time_constant;
 
-  if (*speed == -1)
+  if (*speed == -1)		/* Requesting current speed */
     {
       *speed = dsp.speed;
       return (ESUCCESS);
     }
   
-  if (*speed < DSP_MIN_SPEED || *speed > DSP_MAX_SPEED)
-    {
-      DPRINTF (("Attempt to set invalid speed (%d)\n", *speed));
-      return (EINVAL);
-    }
+  if (*speed < DSP_MIN_SPEED)
+    *speed = DSP_MIN_SPEED;
+  if (*speed > DSP_MAX_SPEED)
+    *speed = DSP_MAX_SPEED;
 
   if (*speed > MAX_LOW_SPEED)
     {
@@ -1164,7 +1348,7 @@ dsp_set_speed (int *speed)
    * buffer should take.
    */
   *speed = dsp.speed;
-  dsp.timeout = DSP_BUF_SIZE * HZ * 2 / dsp.speed;
+  COMPUTE_DSP_TIMEOUT;
   DPRINTF (("Speed set to %d.\n", dsp.speed));
   return (ESUCCESS);
 }
@@ -1193,7 +1377,7 @@ dsp_set_compression (int mode)
   switch (mode)
     {
     case ADPCM_4:
-    case ADPCM_2_6:
+    case ADPCM_3:
     case ADPCM_2:
       if (dsp.hispeed)
 	return (EINVAL);      /* Fall through.. */
@@ -1221,6 +1405,10 @@ dsp_command (BYTE val)
 #endif
   for (i = DSP_LOOP_MAX; i; i--)
     {
+#ifdef SB20			/* SB DSP 2.0 needs these? */
+      tenmicrosec ();
+      tenmicrosec ();
+#endif
       if ((inb (DSP_STATUS) & DSP_BUSY) == 0)
 	{
 	  outb(DSP_COMMAND, val);
@@ -1232,6 +1420,34 @@ dsp_command (BYTE val)
   return (FAIL);
 }
 
+/*
+ * Sets the DSP buffer size.
+ * Smaller sizes make the DSP more responsive,
+ * but require more CPU attention.
+ * Note that the size must be an even value for stereo to work.
+ * We just reject odd sizes.
+ * -1 is magic, it is replaced with the current buffer size.
+ */
+static int
+dsp_set_bufsize (int *newsize)
+{
+  if (*newsize == -1)
+    {
+      *newsize = dsp.bufsize;
+      return (ESUCCESS);
+    }
+  if (*newsize < DSP_MIN_BUFSIZE || *newsize > DSP_MAX_BUFSIZE
+      || *newsize & 1)
+    {
+      DPRINTF (("sb: Attempt to set dsp.bufsize to %d!\n", *newsize));
+      return (EINVAL);
+    }
+  
+  dsp.bufsize = *newsize;
+  COMPUTE_DSP_TIMEOUT;
+  return (ESUCCESS);
+}
+
 /*
  * This turns stereo playback/recording on and off.
  * For playback, it seems to only be a bit in the mixer.
@@ -1271,7 +1487,7 @@ fm_ioctl (int cmd, caddr_t arg)
 }
 
 static int
-fm_open (int flags)
+fm_open (void)
 {
   if (status.fm_in_use)
     return (EBUSY);
@@ -1302,7 +1518,7 @@ fm_reset (void)
 {
   int reg;
   for (reg = 1; reg <= 0xf5; reg++)
-    fm_send (BOTH, reg, 0);
+    fm_send (FM_BOTH, reg, 0);
 
   return (ESUCCESS);;
 }
@@ -1322,13 +1538,13 @@ fm_send (int channel, unsigned char reg, unsigned char val)
 #endif
   switch (channel)
     {
-    case LEFT:
+    case FM_LEFT:
       sb_sendb (FM_ADDR_L, reg, FM_DATA_L, val);
       break;
-    case RIGHT:
+    case FM_RIGHT:
       sb_sendb (FM_ADDR_R, reg, FM_DATA_R, val);
       break;
-    case BOTH:
+    case FM_BOTH:
       sb_sendb (FM_ADDR_B, reg, FM_DATA_B, val);
       break;
     default:
@@ -1358,8 +1574,8 @@ fm_play_note (struct sb_fm_note *n)
   if (voice < 0 || voice > 8)
     return (EINVAL);
 
-  ch = n->channel;		/* LEFT, RIGHT, or BOTH */
-  if (ch != LEFT && ch != RIGHT && ch != BOTH)
+  ch = n->channel;		/* FM_LEFT, FM_RIGHT, or FM_BOTH */
+  if (ch != FM_LEFT && ch != FM_RIGHT && ch != FM_BOTH)
     return (EINVAL);
 
   /*
@@ -1396,7 +1612,7 @@ fm_play_note (struct sb_fm_note *n)
 static int
 fm_set_params (struct sb_fm_params *p)
 {
-  if (p->channel != LEFT && p->channel != RIGHT && p->channel != BOTH)
+  if (p->channel != FM_LEFT && p->channel != FM_RIGHT && p->channel != FM_BOTH)
     return (EINVAL);
   
   fm_send (p->channel, DEPTHS,
@@ -1582,7 +1798,7 @@ mixer_set_levels (struct sb_mixer_levels *l)
       || l->line.l & ~0xF || l->line.r & ~0xF
       || l->voc.l & ~0xF || l->voc.r & ~0xF
       || l->fm.l & ~0xF || l->fm.r & ~0xF
-      || l->cd & ~0xF
+      || l->cd.l & ~0xF || l->cd.r & ~0xF
       || l->mic & ~0x7)
     return (EINVAL);
 
@@ -1590,7 +1806,7 @@ mixer_set_levels (struct sb_mixer_levels *l)
   mixer_send (VOL_LINE, (l->line.l << 4) | l->line.r);
   mixer_send (VOL_VOC, (l->voc.l << 4) | l->voc.r);
   mixer_send (VOL_FM, (l->fm.l << 4) | l->fm.r);
-  mixer_send (VOL_CD, l->cd);
+  mixer_send (VOL_CD, (l->cd.l << 4) | l->cd.r);
   mixer_send (VOL_MIC, l->mic);
   return (ESUCCESS);
 }
@@ -1648,7 +1864,8 @@ mixer_get_levels (struct sb_mixer_levels *l)
   l->fm.r = B4(val);
   
   val = mixer_read_reg (VOL_CD); /* CD */
-  l->cd = B4(val);
+  l->cd.l = B4(val >> 4);
+  l->cd.r = B4(val);
 
   val = mixer_read_reg (VOL_MIC); /* Microphone */
   l->mic = B3(val);
@@ -1761,7 +1978,7 @@ mixer_reset (void)
   p.hifreq_filter = TRUE;
   p.record_source = SRC_LINE;
 
-  l.cd = 1;
+  l.cd.l = l.cd.r = 0;
   l.mic = 1;
   l.master.l = l.master.r = 11;
   l.line.l = l.line.r = 15;
@@ -1803,12 +2020,14 @@ mixer_read_reg (BYTE reg)
  * MIDI functions
  */
 static int
-midi_open (int flags)
+midi_open (void)
 {
   if (status.dsp_in_use)
     return (EBUSY);
 
   status.dsp_in_use = TRUE;
+  status.selects[SB_MIDI_NUM] = NULL;
+  status.sel_coll[SB_MIDI_NUM] = FALSE;
   dsp_reset ();			/* Resets card and inits variables */
 
   /*
@@ -1837,6 +2056,28 @@ midi_close ()
     return (ESRCH);
 }
 
+static int
+midi_select (int rw)
+{
+  int s = splhigh();
+  DPRINTF (("sb: midi_select() called with rw = %d.\n", rw));
+  if (((rw == SELECT_WRITE))
+      || ((rw == SELECT_READ) && (inb (DSP_RDAVAIL) & DSP_DATA_AVAIL)))
+    {
+      splx (s);
+      return (1);		/* Give the go-ahead. */
+    }
+
+  /* It's not ready yet, so put them on the list of procs to notify */
+  if (status.selects[SB_MIDI_NUM])
+    status.sel_coll[SB_MIDI_NUM] = TRUE;
+  else
+    status.selects[SB_MIDI_NUM] = (struct proc *) current_thread();
+  
+  splx (s);
+  return 0;
+}
+
 static int
 midi_read (struct uio *uio)
 {
@@ -1855,10 +2096,11 @@ midi_read (struct uio *uio)
 	{
 	  while (!(inb (DSP_RDAVAIL) & DSP_DATA_AVAIL))
 	    {
-	      if (dsp.dont_block) /* Maybe they don't want to wait.. */
+	      /* Maybe they don't want to wait.. */
+	      if (status.flags[SB_MIDI_NUM] & FNDELAY)
 		{
 		  splx (s);
-		  return (ESUCCESS);
+		  return (EWOULDBLOCK);
 		}
 	      wait_for_intr (DSP_UNIT, 0, PSLEP); /* Maybe they do. */
 	    }
@@ -1913,11 +2155,19 @@ midi_write (struct uio *uio)
   return (ESUCCESS);
 }
 
+/*
+ * No valid ioctls for MIDI yet, but I have some ideas.
+ */
 static int
 midi_ioctl (int cmd, caddr_t arg)
 {
+#if 0
   switch (cmd)
     {
+    default:
+      return (EINVAL);
+    }
+#endif
       /*
        * Shoot.  I can't figure this out, and I don't have any examples.
        * This is supposed to support the fcntl() call.
@@ -1934,9 +2184,6 @@ midi_ioctl (int cmd, caddr_t arg)
       dsp.dont_block = arg->sg_flags & FNDELAY;
       break;
 #endif
-    default:
-      return (EINVAL);
-    }
   return (ESUCCESS);
 }
 
